from json_database import JsonStorage
from os.path import join, exists, isdir
from os import makedirs
from jarbas_hive_mind.settings import DEFAULT_PORT, DATA_PATH, CERTS_PATH, \
    MYCROFT_WEBSOCKET_CONFIG


_DEFAULT_CONFIG = join(DATA_PATH, "HiveMind.conf")


def default_config():
    default = JsonStorage(_DEFAULT_CONFIG)
    default["max_connections"] = -1
    default["port"] = DEFAULT_PORT
    default["data_path"] = DATA_PATH
    default["ssl"] = {
        "certificates": CERTS_PATH,
        "ssl_certfile": "HiveMind.crt",
        "ssl_keyfile": "HiveMind.key"
    }
    default["database"] = join(DATA_PATH, "database", "clients.db")

    default["log_blacklist"] = []
    default["mycroft_bus"] = MYCROFT_WEBSOCKET_CONFIG

    return default


if not exists(_DEFAULT_CONFIG):
    CONFIGURATION = default_config()
    CONFIGURATION.store()
else:
    CONFIGURATION = JsonStorage(_DEFAULT_CONFIG)

# ensure directories exist
if not isdir(CONFIGURATION["data_path"]):
    makedirs(CONFIGURATION["data_path"])

if not isdir(CONFIGURATION["ssl"]["certificates"]):
    makedirs(CONFIGURATION["ssl"]["certificates"])


import base64
from autobahn.twisted.websocket import WebSocketServerProtocol, \
    WebSocketServerFactory
from jarbas_hive_mind.database import ClientDatabase
from jarbas_hive_mind.exceptions import UnauthorizedKeyError
from jarbas_utils.log import LOG
from jarbas_utils.messagebus import Message, get_mycroft_bus
from jarbas_hive_mind.utils import decrypt_from_json, encrypt_as_json
from jarbas_hive_mind.interface import HiveMindMasterInterface
import json
from jarbas_hive_mind.utils import get_ip
from jarbas_hive_mind.discovery.ssdp import SSDPServer
from jarbas_hive_mind.discovery.upnp_server import UPNPHTTPServer
from jarbas_hive_mind.discovery.zero import ZeroConfAnnounce
import uuid

platform = "HiveMindV0.7"


# protocol
class HiveMindProtocol(WebSocketServerProtocol):

    @staticmethod
    def decode_auth(request):
        auth = request.headers.get("authorization")
        if not auth:
            cookie = request.headers.get("cookie")
            if cookie:
                auth = cookie.replace("X-Authorization=", "")
                userpass_encoded = bytes(auth, encoding="utf-8")
        else:
            userpass_encoded = bytes(auth, encoding="utf-8")
            if userpass_encoded.startswith(b"Basic "):
                userpass_encoded = userpass_encoded[6:-2]
            else:
                userpass_encoded = userpass_encoded[2:-1]

        userpass_decoded = base64.b64decode(userpass_encoded).decode("utf-8")
        name, key = userpass_decoded.split(":")
        return name, key

    def onConnect(self, request):

        LOG.info("Client connecting: {0}".format(request.peer))

        name, key = self.decode_auth(request)

        ip = request.peer.split(":")[1]
        context = {"source": self.peer}
        self.platform = request.headers.get("platform", "unknown")

        try:
            with ClientDatabase() as users:
                user = users.get_client_by_api_key(key)
                if not user:
                    raise UnauthorizedKeyError
                self.crypto_key = users.get_crypto_key(key)
        except UnauthorizedKeyError:
            LOG.error("Client provided an invalid api key")
            self.factory.mycroft_send("hive.client.connection.error",
                                      {"error": "invalid api key",
                                       "ip": ip,
                                       "api_key": key,
                                       "platform": self.platform},
                                      context)
            raise

        # send message to internal mycroft bus
        data = {"ip": ip, "headers": request.headers}
        with ClientDatabase() as users:
            self.blacklist = users.get_blacklist_by_api_key(key)
        self.factory.mycroft_send("hive.client.connect", data, context)
        # return a pair with WS protocol spoken (or None for any) and
        # custom headers to send in initial WS opening handshake HTTP response
        headers = {"server": platform}
        return (None, headers)

    def onOpen(self):
        """
       Connection from client is opened. Fires after opening
       websockets handshake has been completed and we can send
       and receive messages.

       Register client in factory, so that it is able to track it.
       """
        self.factory.register_client(self, self.platform)
        LOG.info("WebSocket connection open.")

    def onMessage(self, payload, isBinary):
        if isBinary:
            LOG.debug(
                "Binary message received: {0} bytes".format(len(payload)))
        else:
            payload = self.decode(payload)
            # LOG.debug(
            #    "Text message received: {0}".format(payload))

        self.factory.on_message(self, payload, isBinary)

    def onClose(self, wasClean, code, reason):
        self.factory.unregister_client(self, reason="connection closed")
        LOG.info("WebSocket connection closed: {0}".format(reason))
        ip = self.peer.split(":")[1]
        data = {"ip": ip, "code": code, "reason": "connection closed",
                "wasClean": wasClean}
        context = {"source": self.peer}
        self.factory.mycroft_send("hive.client.disconnect", data, context)

    def connectionLost(self, reason):
        """
       Client lost connection, either disconnected or some error.
       Remove client from list of tracked connections.
       """
        self.factory.unregister_client(self, reason="connection lost")
        LOG.info("WebSocket connection lost: {0}".format(reason))
        ip = self.peer.split(":")[1]
        data = {"ip": ip, "reason": "connection lost"}
        context = {"source": self.peer}
        self.factory.mycroft_send("hive.client.disconnect", data, context)

    def decode(self, payload):
        payload = payload.decode("utf-8")
        if self.crypto_key:
            payload = decrypt_from_json(self.crypto_key, payload)
        return payload

    def sendMessage(self,
                    payload,
                    isBinary=False,
                    fragmentSize=None,
                    sync=False,
                    doNotCompress=False):
        if isinstance(payload, dict):
            payload = json.dumps(payload)
        if self.crypto_key and not isBinary:
            payload = encrypt_as_json(self.crypto_key, payload)
        if isinstance(payload, str):
            payload = bytes(payload, encoding="utf-8")
        super().sendMessage(payload, isBinary,
                            fragmentSize=fragmentSize,
                            sync=sync,
                            doNotCompress=doNotCompress)


class HiveMind(WebSocketServerFactory):
    def __init__(self, bus=None, announce=True, *args, **kwargs):
        super(HiveMind, self).__init__(*args, **kwargs)
        # list of clients
        self.listener = None
        self.clients = {}
        # ip block policy
        self.ip_list = []
        self.blacklist = True  # if False, ip_list is a whitelist
        # mycroft_ws
        self.bus = bus or get_mycroft_bus()
        self.register_mycroft_messages()

        self.interface = HiveMindMasterInterface(self)
        self.announce = announce
        self.upnp_server = None
        self.ssdp = None
        self.zero = None

    def start_announcing(self):
        device_uuid = uuid.uuid4()
        local_ip_address = get_ip()
        hivemind_socket = self.listener.address.replace("0.0.0.0",
                                                        local_ip_address)

        if self.zero is None:
            LOG.info("Registering zeroconf:HiveMind-websocket " +
                     hivemind_socket)
            self.zero = ZeroConfAnnounce(uuid=device_uuid,
                                         port=self.port,
                                         host=hivemind_socket)
            self.zero.daemon = True
            self.zero.start()

        if self.ssdp is None or self.upnp_server is None:
            self.upnp_server = UPNPHTTPServer(8088,
                                              friendly_name="JarbasHiveMind Master",
                                              manufacturer='JarbasAI',
                                              manufacturer_url='https://ai-jarbas.gitbook.io/jarbasai/',
                                              model_description='Jarbas HiveMind',
                                              model_name="HiveMind-core",
                                              model_number="0.9",
                                              model_url="https://github.com/OpenJarbas/HiveMind-core",
                                              serial_number=platform,
                                              uuid=device_uuid,
                                              presentation_url=hivemind_socket,
                                              host=local_ip_address)
            self.upnp_server.start()

            self.ssdp = SSDPServer()
            self.ssdp.register('local',
                               'uuid:{}::upnp:HiveMind-websocket'.format(device_uuid),
                               'upnp:HiveMind-websocket',
                               self.upnp_server.path)
            self.ssdp.start()

    def bind(self, listener):
        self.listener = listener
        if self.announce:
            self.start_announcing()

    @property
    def peer(self):
        if self.listener:
            return self.listener.peer
        return None

    @property
    def node_id(self):
        return self.peer + ":MASTER"

    def mycroft_send(self, type, data=None, context=None):
        data = data or {}
        context = context or {}
        if "client_name" not in context:
            context["client_name"] = platform
        self.bus.emit(Message(type, data, context))

    def register_mycroft_messages(self):
        self.bus.on("message", self.handle_outgoing_mycroft)
        self.bus.on('hive.send', self.handle_send)

    def shutdown(self):
        self.bus.remove('message', self.handle_outgoing_mycroft)
        self.bus.remove('hive.send', self.handle_send)

    # websocket handlers
    def register_client(self, client, platform=None):
        """
       Add client to list of managed connections.
       """
        platform = platform or "unknown"
        LOG.info("registering client: " + str(client.peer))
        t, ip, sock = client.peer.split(":")
        # see if ip address is blacklisted
        if ip in self.ip_list and self.blacklist:
            LOG.warning("Blacklisted ip tried to connect: " + ip)
            self.unregister_client(client, reason="Blacklisted ip")
            return
        # see if ip address is whitelisted
        elif ip not in self.ip_list and not self.blacklist:
            LOG.warning("Unknown ip tried to connect: " + ip)
            #  if not whitelisted kick
            self.unregister_client(client, reason="Unknown ip")
            return
        self.clients[client.peer] = {"instance": client,
                                     "status": "connected",
                                     "platform": platform}

    def unregister_client(self, client, code=3078,
                          reason="unregister client request"):
        """
       Remove client from list of managed connections.
       """

        LOG.info("deregistering client: " + str(client.peer))
        if client.peer in self.clients.keys():
            client_data = self.clients[client.peer] or {}
            j, ip, sock_num = client.peer.split(":")
            context = {"user": client_data.get("names", ["unknown_user"])[0],
                       "source": client.peer}
            self.bus.emit(
                Message("hive.client.disconnect",
                        {"reason": reason, "ip": ip, "sock": sock_num},
                        context))
            client.sendClose(code, reason)
            self.clients.pop(client.peer)

    def on_message(self, client, payload, isBinary):
        """
       Process message from client, decide what to do internally here
       """
        client_protocol, ip, sock_num = client.peer.split(":")

        if isBinary:
            # TODO receive files
            pass
        else:
            # Check protocol
            data = json.loads(payload)
            payload = data["payload"]
            msg_type = data["msg_type"]
            data["source_peer"] = client.peer

            # slave does not know peer name on master, update it
            if data.get("route"):
                data["route"][-1]["source"] = client.peer
                if self.peer not in data["route"][-1]["targets"]:
                    data["route"][-1]["targets"].append(self.peer)
            else:
                data["route"] = [{"source": client.peer,
                                  "targets": [self.peer]}]

            if msg_type == "bus":
                self.handle_bus_message(payload, client)
            elif msg_type == "propagate":
                self.handle_propagate_message(data, client)
            elif msg_type == "broadcast":
                self.handle_broadcast_message(data, client)
            elif msg_type == "escalate":
                self.handle_escalate_message(data, client)

    # HiveMind protocol messages -  from DOWNstream
    def handle_bus_message(self, payload, client):
        # Generate mycroft Message
        if isinstance(payload, str):
            payload = json.loads(payload)
        msg_type = payload.get("msg_type") or payload["type"]
        data = payload.get("data") or {}
        context = payload.get("context") or {}
        message = Message(msg_type, data, context)
        message.context["source"] = client.peer
        message.context["destination"] = "skills"
        self.handle_incoming_mycroft(message, client)

    def handle_broadcast_message(self, data, client):
        # Slaves are not allowed to broadcast, by definition broadcast goes
        # downstream only, use propagate instead
        LOG.debug("Ignoring broadcast message from downstream, illegal action")
        # TODO kick client for misbehaviour so it stops doing that?

    def handle_propagate_message(self, data, client):

        payload = data["payload"]

        LOG.info("Received propagate message at: " + self.node_id)
        LOG.debug("ROUTE: " + str(data["route"]))
        LOG.debug("PAYLOAD: " + str(payload))

        self.interface.propagate(payload, data)

    def handle_escalate_message(self, data, client):
        payload = data["payload"]

        LOG.info("Received escalate message at: " + self.node_id)
        LOG.debug("ROUTE: " + str(data["route"]))
        LOG.debug("PAYLOAD: " + str(payload))

        # TODO Try to answer

        # else escalate again
        self.interface.escalate(payload, data)

    # parsed protocol messages
    def handle_incoming_mycroft(self, message, client):
        # A Slave wants to inject a message in internal mycroft bus
        # You are a Master, authorize bus message

        client_protocol, ip, sock_num = client.peer.split(":")

        # messages/skills/intents per user
        if message.msg_type in client.blacklist.get("messages", []):
            LOG.warning(client.peer + " sent a blacklisted message "
                                      "type: " + message.msg_type)
            return
        # TODO check intent / skill that will trigger

        # send client message to internal mycroft bus
        LOG.info("Forwarding message to mycroft bus from client: " +
                 str(client.peer))
        self.mycroft_send(message.msg_type, message.data, message.context)

    def handle_client_bus(self, message, client):
        # this message is going inside the client bus
        # take any metrics you need
        LOG.info("Monitoring bus from client: " + client.peer)
        assert isinstance(message, Message)

    # mycroft handlers
    def handle_send(self, message):
        payload = message.data.get("payload")
        peer = message.data.get("peer")
        msg_type = message.data["msg_type"]
        if msg_type == "propagate":
            self.interface.propagate(payload, message.data)

        elif msg_type == "broadcast":
            # slaves can not broadcast and will send a bus message instead
            # if the mycroft device has it's own hive the broadcast is
            # handled here
            self.interface.broadcast(payload, message.data)

        elif msg_type == "escalate":
            # only slaves can escalate, ignore silently
            pass
        # NOT a protocol specific message, send directly to requested peer
        elif peer:
            if peer in self.clients:
                # send message to client
                client = self.clients[peer].get("instance")
                self.interface.send(payload, client)
            else:
                LOG.error("That client is not connected")
                self.mycroft_send("hive.client.send.error",
                                  {"error": "That client is not connected",
                                   "peer": peer}, message.context)

    def handle_outgoing_mycroft(self, message=None):
        # forward internal messages to clients if they are the target
        if isinstance(message, dict):
            message = json.dumps(message)
        if isinstance(message, str):
            message = Message.deserialize(message)
        if message.msg_type == "complete_intent_failure":
            message.msg_type = "hive.complete_intent_failure"
        message.context = message.context or {}
        peer = message.context.get("destination")
        if peer and peer in self.clients:
            client = self.clients[peer].get("instance")
            payload = {"msg_type": "bus",
                       "payload": message.serialize()
                       }
            self.interface.send(payload, client)



import upnpclient
import threading
from time import sleep
from jarbas_utils.log import LOG
from jarbas_utils.xml_helper import xml2dict
from jarbas_hive_mind.slave import HiveMindSlave
from jarbas_hive_mind.slave.terminal import HiveMindTerminal
from jarbas_hive_mind.discovery.zero import ZeroScanner
import requests


class _Device:
    def __init__(self, host, device_type='HiveMind-websocket'):
        self.host = host
        self.device_type = device_type

    @property
    def services(self):
        return {}

    @property
    def location(self):
        return None

    @property
    def device_name(self):
        return self.host

    @property
    def friendly_name(self):
        return self.device_name

    @property
    def model_description(self):
        return self.device_name

    @property
    def model_name(self):
        return self.device_type

    @property
    def model_number(self):
        return "0.1"

    @property
    def udn(self):
        return self.model_name + ":" + self.model_number

    @property
    def address(self):
        return self.location

    @property
    def data(self):
        return {"host": self.host,
                "type": self.device_type}


class HiveMindNode:
    def __init__(self, d=None):
        self.device = d
        self._data = None

    @property
    def services(self):
        return self.device.service_map

    @property
    def xml(self):
        return self.device.location

    @property
    def device_name(self):
        return self.device.device_name

    @property
    def friendly_name(self):
        return self.device.friendly_name

    @property
    def description(self):
        return self.device.model_description

    @property
    def node_tyoe(self):
        return self.device.model_name

    @property
    def version(self):
        return self.device.model_number

    @property
    def device_id(self):
        return self.device.udn

    @property
    def data(self):
        if self.xml and self._data is None:
            LOG.info("Fetching Node data: {url}".format(url=self.xml))
            xml = requests.get(self.xml).text
            self._data = xml2dict(xml)
        return self._data

    @property
    def address(self):
        try:
            if self.device.location:
                services = self.data["root"]["device"]['serviceList']
                for s in services:
                    service = services[s]
                    if service["serviceType"] == \
                            'urn:jarbasAi:HiveMind:service:Master':
                        return service["URLBase"]
        except:
            pass
        return self.device.data.get("host")

    @property
    def host(self):
        return ":".join(self.address.split(":")[:-1])

    @property
    def port(self):
        return int(self.address.split(":")[-1])

    def connect(self, headers, crypto_key=None, bus=None,
                node_type=None):
        try:
            # TODO cyclic import
            from jarbas_hive_mind import HiveMindConnection
            con = HiveMindConnection(self.host, self.port)
            if node_type:
                clazz = node_type
            else:
                if bus:
                    clazz = HiveMindSlave
                else:
                    clazz = HiveMindTerminal
            if bus:
                component = clazz(bus=bus,
                                  headers=headers,
                                  crypto_key=crypto_key)
            else:
                component = clazz(headers=headers,
                                  crypto_key=crypto_key)
            # will check url for ssl
            LOG.debug("Connecting to HiveMind websocket @ {url}".format(
                url=self.address))
            con.connect(component)
        except Exception as e:
            LOG.error("Connection failed")
            LOG.exception(e)


class LocalDiscovery(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pause = 20  # scan every 20 seconds
        self._nodes = {}
        self.blacklist = []
        self.running = False
        self.connected = False
        self.zero = ZeroScanner()
        self.zero.on_new_node = self.on_new_zeroconf_node
        self.zero.daemon = True
        self.zero.start()

    def on_new_zeroconf_node(self, node):
        node = HiveMindNode(_Device(node["address"]))
        LOG.info("ZeroConf Node Found: " + str(node.address))
        self._nodes[node.address] = node
        self.on_new_node(node)

    def on_new_upnp_node(self, node):
        LOG.info("UpNp Node Found: " + node.xml)
        self._nodes[node.xml] = node
        self.on_new_node(node)

    def on_new_node(self, node):
        pass

    @property
    def nodes(self):
        return self._nodes

    def scan(self):
        for node_url in self._nodes:
            if node_url in self.blacklist:
                continue
            yield node_url
        for node_url in self.upnp_scan():
            if node_url in self.blacklist:
                continue
            self.blacklist.append(node_url)
            yield node_url

    def upnp_scan(self):
        devices = upnpclient.discover()
        for d in devices:
            if d.location in self.nodes:
                continue
            if d.model_name == "HiveMind-core":
                node = HiveMindNode(d)
                self.on_new_upnp_node(node)
                yield node.xml

    def run(self) -> None:
        self.running = True
        while self.running:
            self.scan()
            sleep(self.pause)
        self.stop()

    def stop(self):
        self.running = False
        self.zero.stop()

    def search_and_connect(self, *args, **kwargs):
        while True:
            # allow zeroconf to get result before upnp check
            sleep(0.5)
            if self.connected:
                continue
            for node_url in self.scan():
                if node_url in self.blacklist:
                    continue

                self.blacklist.append(node_url)
                node = self.nodes[node_url]
                node.connect(*args, **kwargs)
                self.connected = True




import json
from jarbas_utils.messagebus import Message


class HiveMindSlaveInterface:
    def __init__(self, connection):
        self.client = connection

    @property
    def bus(self):
        # if bus is None -> Terminal
        # else -> Slave
        return self.client.bus

    @property
    def node_id(self):
        return self.client.node_id

    @property
    def peer(self):
        return self.client.peer

    def send(self, payload):
        if isinstance(payload, dict):
            payload = json.dumps(payload)
        if not isinstance(payload, bytes):
            payload = bytes(payload, encoding="utf-8")

        self.client.sendMessage(payload)

    def send_to_hivemind_bus(self, payload):
        if isinstance(payload, Message):
            payload = payload.serialize()
        payload = {"msg_type": "bus",
                   "payload": payload,
                   "node": self.node_id,
                   "source_peer": self.peer
                   }
        self.send(payload)

    def broadcast(self, payload, msg_data=None):
        msg_data = msg_data or {}
        payload = {"msg_type": "broadcast",
                   "payload": payload,
                   "route": msg_data.get("route", []),
                   "source_peer": self.peer,
                   "node": self.client.node_id
                   }
        if self.bus:
            message = Message("hive.send",
                              payload,
                              {"destination": "hive",
                               "source": self.peer})
            self.bus.emit(message)

    def propagate(self, payload, msg_data=None):
        msg_data = msg_data or {}
        payload = {"msg_type": "propagate",
                   "payload": payload,
                   "route": msg_data.get("route", []),
                   "source_peer": self.peer,
                   "node": self.client.node_id
                   }

        no_send = [n["source"] for n in msg_data.get("route", [])]
        if self.peer not in no_send:
            self.send(payload)

        if self.bus and msg_data.get("node", "") == \
                self.peer + ":MASTER":
            message = Message("hive.send",
                              payload,
                              {"destination": "hive",
                               "source": self.peer})
            self.bus.emit(message)

    def escalate(self, payload, msg_data=None):
        msg_data = msg_data or {}
        payload = {"msg_type": "escalate",
                   "payload": payload,
                   "route": msg_data.get("route", []),
                   "source_peer": self.peer,
                   "node": self.client.node_id
                   }

        route_data = {"source": self.peer,
                      # when master receives this it will update this field
                      "targets": [self.peer]}
        if route_data not in payload["route"]:
            payload["route"].append(route_data)
        self.send(payload)

    def query(self, payload, msg_data=None):
        raise NotImplementedError

    def cascade(self, payload, msg_data=None):
        raise NotImplementedError


class HiveMindMasterInterface:
    def __init__(self, listener):
        self.listener = listener

    @property
    def peer(self):
        return self.listener.peer

    @property
    def clients(self):
        return [c["instance"] for c in
                self.listener.clients.values()]

    @property
    def node_id(self):
        return self.listener.node_id

    @property
    def bus(self):
        return self.listener.bus

    def send(self, payload, client):
        if isinstance(client, str):
            client = self.listener.clients[client]
        if isinstance(payload, dict):
            payload = json.dumps(payload)
        if not isinstance(payload, bytes):
            payload = bytes(payload, encoding="utf-8")
        client.sendMessage(payload)

    def send_to_many(self, payload, no_send=None):
        no_send = no_send or []
        payload["route"] = payload["route"] or []
        route_data = {"source": self.peer,
                      "targets": [c.peer for c in self.clients]}
        # No resend
        if route_data in payload["route"]:
            return

        payload["route"].append(route_data)

        for client in self.clients:
            if client.peer not in no_send:
                self.send(payload, client)

    def broadcast(self, payload, msg_data=None):
        msg_data = msg_data or {}
        payload = {"msg_type": "broadcast",
                   "payload": payload,
                   "route": msg_data.get("route", []),
                   "source_peer": self.peer,
                   "node": self.node_id
                   }
        no_send = [n["source"] for n in msg_data.get("route", [])]
        self.send_to_many(payload, no_send)

    def propagate(self, payload, msg_data=None):
        msg_data = msg_data or {}
        payload = {"msg_type": "propagate",
                   "payload": payload,
                   "route": msg_data.get("route", []),
                   "source_peer": self.peer,
                   "node": self.node_id
                   }

        no_send = [n["source"] for n in msg_data.get("route", [])]
        self.send_to_many(payload, no_send)

        # tell Slave to propagate upstream
        if self.bus and msg_data.get("source_peer") != self.peer:
            payload["msg_type"] = "propagate"
            message = Message("hive.send",
                              payload,
                              {"destination": "hive",
                               "source": self.peer})
            self.bus.emit(message)

    def escalate(self, payload, msg_data=None):
        msg_data = msg_data or {}
        payload = {"msg_type": "escalate",
                   "payload": payload,
                   "route": msg_data.get("route", []),
                   "source_peer": self.peer,
                   "node": self.node_id
                   }
        # tell Slave to escalate upstream
        if self.bus and msg_data.get("source_peer") != self.peer:
            message = Message("hive.send",
                              payload,
                              {"destination": "hive",
                               "source": self.peer})
            self.bus.emit(message)

    def query(self, payload, msg_data=None):
        raise NotImplementedError

    def cascade(self, payload, msg_data=None):
        raise NotImplementedError




from jarbas_hive_mind.slave.terminal import HiveMindTerminal, \
    HiveMindTerminalProtocol
from jarbas_utils.log import LOG
from jarbas_utils.messagebus import Message, get_mycroft_bus
import json

platform = "HiveMindSlaveV0.2"


class HiveMindSlaveProtocol(HiveMindTerminalProtocol):

    @property
    def bus(self):
        return self.factory.bus

    def onConnect(self, response):
        self.bus.emit(Message("hive.mind.connected",
                              {"server_id": response.headers["server"]}))
        super().onConnect(response)

    def onOpen(self):
        self.bus.emit(Message("hive.mind.websocket.open"))
        super().onOpen()

    def onClose(self, wasClean, code, reason):
        self.bus.emit(Message("hive.mind.client.closed",
                              {"wasClean": wasClean,
                               "reason": reason,
                               "code": code}))
        super().onClose(wasClean, code, reason)


class HiveMindSlave(HiveMindTerminal):
    protocol = HiveMindSlaveProtocol

    def __init__(self, bus=None, *args, **kwargs):
        super(HiveMindSlave, self).__init__(*args, **kwargs)
        # mycroft_ws
        self.bus = bus or get_mycroft_bus()
        self.register_mycroft_messages()

    # mycroft methods
    def register_mycroft_messages(self):
        self.bus.on("message", self.handle_outgoing_mycroft)
        self.bus.on("hive.send", self.handle_send)

    def shutdown(self):
        self.bus.remove("message", self.handle_outgoing_mycroft)
        self.bus.remove("hive.send", self.handle_send)

    def handle_send(self, message):
        payload = message.data
        msg_type = payload["msg_type"]

        if msg_type == "bus":
            mycroft_message = payload["payload"]
            self.send_to_hivemind_bus(mycroft_message)
        elif msg_type == "propagate":
            self.interface.propagate(payload["payload"], payload)
        elif msg_type == "broadcast":
            # LOG.debug("Slaves can not broadcast messages")
            # Ignore silently, if a Master is connected to bus it will
            # handle it
            pass
        elif msg_type == "escalate":
            self.interface.escalate(payload["payload"], payload)
        else:
            LOG.error("Unknown HiveMind protocol msg_type")

    def handle_outgoing_mycroft(self, message=None):
        # forward internal messages to clients if they are the target
        if isinstance(message, dict):
            message = json.dumps(message)
        if isinstance(message, str):
            message = Message.deserialize(message)
        if message.msg_type == "complete_intent_failure":
            message.msg_type = "hive.complete_intent_failure"
        message.context = message.context or {}
        message.context["source"] = self.client.peer
        peer = message.context.get("destination")
        message = message.serialize()
        if peer and peer == self.client.peer:
            payload = {"msg_type": "bus",
                       "payload": message
                       }
            self.interface.send(payload)

    # parsed protocol messages
    def handle_incoming_mycroft(self, payload):
        """ HiveMind is sending a mycroft bus message"""
        # you are a slave_connection, just forward to the bus
        self.bus.emit(Message("hive.message.received", payload))
        super().handle_incoming_message(payload)

    # websocket handlers
    def on_binary(self, payload):
        # TODO receive binary file
        LOG.info("[BINARY MESSAGE]")

    def send_to_hivemind_bus(self, payload):
        super().send_to_hivemind_bus(payload)
        self.bus.emit(Message("hive.message.sent",
                              {"payload": payload}))




from os import makedirs
from os.path import isdir, join, expanduser

DATA_PATH = expanduser("~/jarbasHiveMind")

if not isdir(DATA_PATH):
    makedirs(DATA_PATH)

CERTS_PATH = join(DATA_PATH, "certs")
if not isdir(CERTS_PATH):
    makedirs(CERTS_PATH)


DB_PATH = join(DATA_PATH, "database")
if not isdir(DB_PATH):
    makedirs(DB_PATH)

CLIENTS_DB = "sqlite:///" + join(DB_PATH, "clients.db")

DEFAULT_PORT = 5678
USE_SSL = True

LOG_BLACKLIST = []

MYCROFT_WEBSOCKET_CONFIG = {
    "host": "0.0.0.0",
    "port": 8181,
    "route": "/core",
    "ssl": False
}



from jarbas_hive_mind.utils import get_ip
from jarbas_utils.log import LOG
import time
from zeroconf import Zeroconf, ServiceInfo
import ipaddress
import threading
from uuid import uuid4
from time import sleep

from zeroconf import ServiceBrowser, ServiceStateChange


class ZeroConfAnnounce(threading.Thread):
    def __init__(self,
                 uuid=None,
                 host=None,
                 port=5678,
                 service_type="HiveMind-websocket",
                 name="HiveMind-Node", *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.port = port
        self.service_type = service_type
        self.host = host or get_ip()
        self.uuid = uuid or str(uuid4())

    @property
    def ssl(self):
        return self.host.startswith("wss://") or \
               self.host.startswith("https://")

    def run(self):
        """Start advertising to other devices about the ip address"""
        # Get the local ip address
        ip = get_ip()

        info = ServiceInfo(
            "_http._tcp.local.",
            " - {}._http._tcp.local.".format(self.name),
            addresses=[ipaddress.ip_address(ip).packed],
            port=self.port,
            properties={"type": self.service_type,
                        "ssl": self.ssl,
                        "host": self.host,
                        "port": self.port},
        )

        zeroconf = Zeroconf()
        # Registering service
        zeroconf.register_service(info)
        try:
            while True:
                time.sleep(0.1)
        finally:
            # Unregister service for whatever reason
            zeroconf.unregister_service(info)
            zeroconf.close()


class ZeroScanner(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.zero = Zeroconf()
        self.browser = ServiceBrowser(self.zero, "_http._tcp.local.",
                                      handlers=[self.on_service_state_change])
        self.nodes = {}
        self.running = False

    def get_nodes(self):
        return self.nodes

    def on_new_node(self, node_data):
        self.nodes[node_data["address"]] = node_data

    def on_node_update(self, node_data):
        pass

    def on_service_state_change(self, zeroconf, service_type, name,
                                state_change):

        if state_change is ServiceStateChange.Added or state_change is \
                ServiceStateChange.Updated:
            info = zeroconf.get_service_info(service_type, name)
            node_data = {}
            if info:
                if info.properties:
                    for key, value in info.properties.items():
                        if key == b"type" and value == b"HiveMind-websocket":
                            node_data["address"] = info._properties[b"host"].decode("utf-8")
                            node_data["weight"] = info.weight
                            node_data["priority"] = info.priority
                            node_data["server"] = info.server
                            node_data["type"] = value.decode("utf-8")
                            if state_change is ServiceStateChange.Added:

                                self.on_new_node(node_data)
                            else:
                                #LOG.info("Node Updated: " + str(node_data))
                                self.on_node_update(node_data)
                            print(zeroconf.cache.__dict__)
                else:
                    LOG.debug("  No properties")

            else:
                LOG.debug("  No info")

    def run(self) -> None:
        self.running = True
        while self.running:
            sleep(0.1)
        self.stop()

    def stop(self):
        self.running = False
        self.zero.close()


if __name__ == '__main__':
    z = ZeroScanner()
    z.start()


